# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Model that wraps all classification models of the mmpretrain package
"""

from __future__ import annotations as _annotations_

import argparse
import inspect
import logging
import os
import shlex
import subprocess
import sys
from threading import Thread
from typing import IO, Any, Callable, Dict, List, Optional, Tuple, Union, cast

import torch
import torch.nn
import yaml
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.types import ImageType, MetricInfo
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError
from mlcvzoo_base.api.interfaces import LogProvider, MetricProvider, NetBased, Trainable
from mlcvzoo_base.api.model import ClassificationModel
from mlcvzoo_base.api.structs import Runtime
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from mlcvzoo_base.data_preparation.annotation_handler import AnnotationHandler
from mlcvzoo_base.data_preparation.annotation_writer.txt_annotation_writer import (
    ClassificationAnnotationWriter,
)
from mmengine.config import Config
from mmengine.runner import Runner, load_checkpoint
from mmpretrain.apis import ImageClassificationInferencer

from mlcvzoo_mmpretrain.configuration import MMPretrainConfig, MMPretrainInferenceConfig
from mlcvzoo_mmpretrain.utils import (
    determine_logs_path,
    extract_best_metric,
    extract_logs,
    init_mm_config,
    modify_config,
    parse_logs,
)

logger = logging.getLogger(__name__)


class MMPretrainModel(
    ClassificationModel[MMPretrainConfig, Union[str, ImageType]],
    NetBased[torch.nn.Module, MMPretrainInferenceConfig],
    Trainable,
    MetricProvider,
    LogProvider,
):
    """
    Class for wrapping mmpretrain rotated object detection models
    """

    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[MMPretrainConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
        is_multi_gpu_instance: bool = False,
        runtime: str = Runtime.DEFAULT,
    ) -> None:
        self.net: Optional[torch.nn.Module] = None
        self.inferencer: Optional[ImageClassificationInferencer] = None

        self.yaml_config_path: Optional[str] = from_yaml
        self.is_multi_gpu_instance: bool = is_multi_gpu_instance

        self.__metrics: Dict[str, MetricInfo] = {}
        self.__best_metric: Optional[MetricInfo] = None

        self.configuration: MMPretrainConfig = self.create_configuration(
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        self.configuration.recursive_string_replacement(
            string_replacement_map=string_replacement_map
        )

        # Config path to mmdetection specific .py file is set
        if self.configuration.mm_config.config_path:
            if string_replacement_map:
                self.configuration.mm_config.config_path = modify_config(
                    config_path=self.configuration.mm_config.config_path,
                    string_replacement_map=string_replacement_map,
                )

        mapper = AnnotationClassMapper(
            class_mapping=self.configuration.class_mapping,
            reduction_mapping=self.configuration.inference_config.reduction_class_mapping,
        )

        self.cfg = init_mm_config(mm_config=self.configuration.mm_config, mapper=mapper)
        self._runner: Optional[Runner] = None

        ClassificationModel.__init__(
            self,
            configuration=self.configuration,
            mapper=mapper,
            init_for_inference=init_for_inference,
            runtime=runtime,
        )
        NetBased.__init__(self, net=self.net)
        Trainable.__init__(self)

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[MMPretrainConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> MMPretrainConfig:
        return cast(
            MMPretrainConfig,
            create_basis_configuration(
                configuration_class=MMPretrainConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @property
    def num_classes(self) -> int:
        return self.mapper.num_classes

    def get_classes_id_dict(self) -> Dict[int, str]:
        return self.mapper.annotation_class_id_to_model_class_name_map

    def _init_inference_model(self) -> None:
        self.inferencer = ImageClassificationInferencer(
            model=self.cfg,
            pretrained=self.configuration.inference_config.checkpoint_path,
            device=self.configuration.inference_config.device_string,
        )
        self.net = self.inferencer.model

    def get_checkpoint_filename_suffix(self) -> str:
        return ".pth"

    def get_training_output_dir(self) -> str:
        return self.configuration.train_config.argparse_config.work_dir

    def get_net(self) -> Optional[torch.nn.Module]:
        return self.net

    def _init_training_model(self) -> None:
        """Additional model initialization for training mode.

        Raises:
            RuntimeError: If the learning rate autoscaler is not configured properly.
            ValueError: If the runtime is not supported.
        """
        if self.runtime == Runtime.DEFAULT:
            # load config
            self.cfg.launcher = self.configuration.train_config.argparse_config.launcher
            self.cfg.experiment_name = self.unique_name
            # We always define the work_dir in the configuration file
            self.cfg.work_dir = self.configuration.train_config.argparse_config.work_dir

            # enable automatic-mixed-precision training
            if self.configuration.train_config.argparse_config.amp is True:
                optim_wrapper = self.cfg.optim_wrapper.type
                if optim_wrapper == "AmpOptimWrapper":
                    logger.warning(
                        "AMP training is already enabled in your config.",
                    )
                else:
                    assert optim_wrapper == "OptimWrapper", (
                        "`--amp` is only supported when the optimizer wrapper type is "
                        f"`OptimWrapper` but got {optim_wrapper}."
                    )
                    self.cfg.optim_wrapper.type = "AmpOptimWrapper"
                    self.cfg.optim_wrapper.loss_scale = "dynamic"

            # enable automatically scaling LR
            if self.configuration.train_config.argparse_config.auto_scale_lr:
                if (
                    "auto_scale_lr" in self.cfg
                    and "enable" in self.cfg.auto_scale_lr
                    and "base_batch_size" in self.cfg.auto_scale_lr
                ):
                    self.cfg.auto_scale_lr.enable = True
                else:
                    raise RuntimeError(
                        'Can not find "auto_scale_lr" or '
                        '"auto_scale_lr.enable" or '
                        '"auto_scale_lr.base_batch_size" in your'
                        " configuration file."
                    )
        else:
            raise ValueError(
                f"Initialization for training is not supported for runtime '{self.runtime}'."
            )

    def store(self, checkpoint_path: str) -> None:
        pass

    def restore(self, checkpoint_path: str) -> None:
        if self.net is None:
            raise ValueError(
                "In order to restore a checkpoint, the net attribute has"
                "to be initialized!"
            )

        logger.info(
            "Load model for %s from %s",
            self.unique_name,
            checkpoint_path,
        )

        checkpoint = load_checkpoint(
            self.net,
            checkpoint_path,
            map_location="cpu",
        )

        if "CLASSES" in checkpoint.get("meta", {}):
            self.net.CLASSES = checkpoint["meta"]["CLASSES"]

    def save_reduced_checkpoint(
        self, input_checkpoint_path: str, output_checkpoint_path: str
    ) -> None:
        """
        Saves a reduced version of a stored checkpoint that does not contain optimizer states
        anymore. Therefore, it keeps the weights and meta information of the source checkpoint.

        Args:
            input_checkpoint_path: Path to source checkpoint file
            output_checkpoint_path: Path to where the checkpoint is saved
        """

        runner = MMPretrainModel._build_runner(cfg=self.cfg)

        # Load input checkpoint dict from and extract the metadata for the output checkpoint.
        # Save reduced checkpoint with metadata of full checkpoint to target directory.
        runner.save_checkpoint(
            out_dir=os.path.dirname(output_checkpoint_path),
            filename=os.path.basename(output_checkpoint_path),
            save_optimizer=False,
            save_param_scheduler=False,
            meta=runner.load_checkpoint(filename=input_checkpoint_path)["meta"],
        )

        logger.info(
            "Saved checkpoint from '%s' in a reduced version to '%s'.",
            input_checkpoint_path,
            output_checkpoint_path,
        )

    def __decode_result(self, result: Dict[str, Any]) -> List[Classification]:
        classifications: List[Classification] = []

        _class_name = None
        try:
            if "pred_class" in result:
                _class_name = result["pred_class"]
                class_identifier = ClassIdentifier(
                    class_id=self.mapper.map_annotation_class_name_to_model_class_id(
                        class_name=_class_name
                    ),
                    class_name=self.mapper.map_annotation_class_name_to_model_class_name(
                        class_name=_class_name
                    ),
                )
                classifications.append(
                    Classification(
                        class_identifier=class_identifier,
                        score=float(result["pred_score"]),
                    )
                )
            elif "pred_label" in result:
                class_identifiers = (
                    self.mapper.map_model_class_id_to_output_class_identifier(
                        class_id=int(result["pred_label"])
                    )
                )
                for class_identifier in class_identifiers:
                    classifications.append(
                        Classification(
                            class_identifier=class_identifier,
                            score=float(result["pred_score"]),
                        )
                    )
        except KeyError as key_error:
            logger.debug(
                "Key not preset in mmpretrain result, result will be skipped, error=%s",
                key_error,
            )
        except ClassMappingNotFoundError:
            logger.debug(
                "Could not find a valid class-mapping for class-name '%s'. "
                "Classification will be skipped",
                _class_name,
            )

        return classifications

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[Classification]]:
        if self.runtime == Runtime.DEFAULT:
            if self.inferencer is None:
                raise ValueError(
                    "The 'inferencer' attribute is not initialized, make sure to instantiate with "
                    "init_for_inference=True"
                )
        else:
            raise ValueError(f"runtime={self.runtime} is not supported for inference.")

        return data_item, self.__decode_result(result=self.inferencer(data_item)[0])

    def predict_many(
        self, data_items: List[Union[str, ImageType]]
    ) -> List[Tuple[Union[str, ImageType], List[Classification]]]:

        classifications_list: List[
            Tuple[Union[str, ImageType], List[Classification]]
        ] = []

        if self.runtime == Runtime.DEFAULT:
            if self.inferencer is None:
                raise ValueError(
                    "The 'inferencer' attribute is not initialized, make sure to instantiate with "
                    "init_for_inference=True"
                )
        else:
            raise ValueError(f"runtime={self.runtime} is not supported for inference.")

        results = self.inferencer(data_items)
        for data_item, result in zip(data_items, results):
            classifications_list.append(
                (data_item, self.__decode_result(result=result))
            )

        return classifications_list

    def train(self) -> None:
        if (
            self.configuration.train_config.multi_gpu_config is None
            or "LOCAL_RANK" in os.environ
        ):
            # Don't start a distributed training if called without a multi GPU config,
            # or run as a worker via torch.distributed as indicated by the presence of a local rank
            logger.info("Training model")
            self._train()
        else:
            logger.info("Training distributed model")
            self._train_multi_gpu()

    def _train_multi_gpu(
        self,
    ) -> None:
        """
        Run mmdet multi-gpu/distributed training.

        Returns:
            None
        """
        if self.configuration.train_config.multi_gpu_config is None:
            raise RuntimeError("No multi GPU config provided")

        distributed_model_config_path = os.path.join(
            self.get_training_output_dir(),
            "{}-worker.yaml".format(self.configuration.unique_name),
        )
        os.makedirs(os.path.dirname(distributed_model_config_path), exist_ok=True)
        with open(distributed_model_config_path, "w+") as yaml_file:
            self.configuration.to_yaml(
                yaml_package=yaml, dumper_cls=yaml.Dumper, stream=yaml_file
            )

        command = (
            f"{sys.executable} -u "
            "-m torch.distributed.run "
            f"--nproc_per_node={self.configuration.train_config.multi_gpu_config.nproc_per_node} "
            f"--master_port={self.configuration.train_config.multi_gpu_config.master_port} "
            f"--master_addr={self.configuration.train_config.multi_gpu_config.master_address} "
            f"--nnodes={self.configuration.train_config.multi_gpu_config.nnodes} "
            f"--node_rank={self.configuration.train_config.multi_gpu_config.node_rank} "
            f"{inspect.getfile(self.__class__)} "
            f"{distributed_model_config_path} "
        )

        subproc_env = dict(os.environ.copy())
        subproc_env["LOGLEVEL"] = "INFO"

        logger.debug("Run command: %s", command)
        result = subprocess.Popen(
            shlex.split(command),
            env=subproc_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=False,
        )
        Thread(
            target=self.__pipe_to_log, args=[result.stdout, logging.INFO], daemon=True
        ).start()
        result.wait()

        if result.returncode:
            logger.error(
                "Command '%s' returned with exit code %i", command, result.returncode
            )
            raise RuntimeError(
                f"Distributed training exited with exitcode != 0, "
                f"exitcode: {result.returncode}"
            )

    @staticmethod
    def __pipe_to_log(pipe: IO[bytes], level: int) -> None:
        with pipe:
            for line in iter(pipe.readline, b""):  # b'\n'-separated lines
                logger.log(level, "Subprocess: %r", line.decode("utf-8").strip())

    @staticmethod
    def _build_runner(cfg: Config) -> Runner:
        from mmengine.registry import RUNNERS

        # build the runner from config
        if "runner_type" not in cfg:
            # build the default runner
            runner = Runner.from_cfg(cfg)
        else:
            # build customized runner from the registry
            # if 'runner_type' is set in the cfg
            runner = RUNNERS.build(cfg)

        return runner

    @staticmethod
    def __generate_txt_annotations(
        annotation_handler: AnnotationHandler,
        output_file_path: str,
        data_root_dir: str,
        seperator: str,
    ) -> None:

        annotations = annotation_handler.parse_training_annotations()

        annotation_writer = ClassificationAnnotationWriter(
            output_file_path=output_file_path,
            data_root_dir=data_root_dir,
            seperator=seperator,
        )

        annotation_writer.write(annotations=annotations)

    def __update_training_data(self) -> None:

        dataset_config = self.configuration.train_config.dataset_config

        if dataset_config is None:
            return

        for context, config in zip(
            ("train", "test", "val"),
            (
                dataset_config.train_annotation_handler_config,
                dataset_config.test_annotation_handler_config,
                dataset_config.val_annotation_handler_config,
            ),
        ):
            if config is not None:
                self.__update_dataloader(
                    context=context,
                    annotation_handler_config=config,
                )
        logger.info("Updated training data")

    def __update_dataloader(
        self,
        context: str,
        annotation_handler_config: AnnotationHandlerConfig,
    ) -> None:
        data_root = self.get_training_output_dir()

        output_file_path = os.path.join(data_root, f"{context}_annotations.txt")

        dataloader = f"{context}_dataloader"
        logger.info(
            f"Update mmpretrain {dataloader} with annotations from "
            f"configuration.train_config.dataset_config.{context}_annotation_handler_config"
        )

        if self.cfg[dataloader] is None:
            return

        self.cfg[dataloader].dataset.ann_file = output_file_path
        self.cfg[dataloader].dataset.data_root = data_root

        # necessary, otherwise the cropped image paths are not found
        self.cfg[dataloader].dataset["data_prefix"] = {
            "img_path": self.get_training_output_dir()
        }

        MMPretrainModel.__generate_txt_annotations(
            annotation_handler=AnnotationHandler(
                mapper=self.mapper,
                configuration=annotation_handler_config,
            ),
            output_file_path=output_file_path,
            data_root_dir=data_root,
            seperator=" ",  # mmpretrain does not support configuration of separator
        )

    def _train(
        self,
    ) -> None:
        self.__update_training_data()

        self._runner = MMPretrainModel._build_runner(cfg=self.cfg)

        # start training
        self._runner.train()

    @staticmethod
    def run(model_class: Callable) -> None:  # type: ignore[type-arg]
        """
        Run training with given model subclass, reading a config from file.
        This function is intended to be called during torch.distributed.run via subclass main
        functions. It expects a single command line parameter called "yaml_config_path" pointing
        to a valid YAML configuration file for the given model subclass.

        Args:
            model_class: the subclass of model.MMPretrainModel to train
        """
        parser: argparse.ArgumentParser = argparse.ArgumentParser()
        parser.add_argument(
            "yaml_config_path",
            help="Load all parameters from the given yaml configuration file",
            type=str,
        )

        args = parser.parse_args()

        mmdet_model = model_class(
            from_yaml=args.yaml_config_path,
            init_for_inference=False,
        )
        mmdet_model.train()

    def determine_training_metrics(
        self,
    ) -> Tuple[Optional[MetricInfo], Dict[str, MetricInfo]]:
        """
        Implement interface for determining the training metrics.
        During the training of a mmpretrain model metrics will be
        logged to a file. This function reads the logs via
        determine_training_logs(...) and extract the best metric
        in the desired format of the MetricProvider interface

        Returns:
            The best metric and its history in stated format.
        """

        if (
            self.configuration.train_config.best_metric_name is not None
            and self._runner is not None
        ):
            return extract_best_metric(
                work_dir=self._runner.cfg.work_dir,
                filename_template=self._runner.cfg.default_hooks.checkpoint.filename_tmpl,
                logs=self.determine_training_logs(),
                best_metric_name=self.configuration.train_config.best_metric_name,
            )

        return None, {}

    def determine_training_logs(
        self,
    ) -> Dict[int, Dict[str, Any]]:
        """
        Determine logs with
        1st Key: Interval
        2nd Key: Metric name
        value: Value of the Metric

        Returns:
            The determined trainings logs
        """

        if self._runner:
            logs_path = determine_logs_path(work_dir=self._runner.cfg.work_dir)

            if logs_path is not None:
                return extract_logs(logs=parse_logs(file_path=logs_path))

        return {}


if __name__ == "__main__":
    MMPretrainModel.run(MMPretrainModel)
