# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Definition of the MMPretrainConfig that is used to configure the MMPretrainModel.
"""

import logging
from typing import Any, Dict, Optional

import related
from attr import define
from config_builder import BaseConfigClass
from mlcvzoo_base.api.configuration import InferenceConfig, ModelConfiguration
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import ReductionMappingConfig

logger = logging.getLogger(__name__)


@define
class MMPretrainTrainArgparseConfig(BaseConfigClass):
    """Argparse parameter from mmpretrain/tools/train.py

    ======================================================================================
    Parameters of the mmpretrain/tools/train.py that are not used in mlcvzoo-mmpretrain:

    If specify checkpoint path, resume from it, while if not
    specify, try to auto resume from the latest checkpoint
    in the work directory:
    resume: str

    NOTE: The local_rank parameter will be set via the argparse argument
          in the training scripts.
    When using PyTorch version >= 2.0.0, the `torch.distributed.launch`
    will pass the `--local-rank` parameter to `tools/train.py` instead
    of `--local_rank`:

    local_rank: int

    cfg_options: Optional[Dict[str, Any]], parameter has been moved to mmconfig TODO

    config: Optional[str], parameter is used via mmconfig TODO
    """

    __related_strict__ = True
    # Gathering of all argparse parameter MMPretrain/tools/train.py:

    # the dir to save logs and models
    work_dir: str = related.StringField()
    # enable automatic-mixed-precision training
    amp: bool = related.BooleanField(default=False)
    # enable automatically scaling LR
    auto_scale_lr: bool = related.BooleanField(required=False, default=False)
    # job launcher
    launcher: str = related.StringField(default="none", required=False)

    def check_values(self) -> bool:
        return self.launcher in ["none", "pytorch", "slurm", "mpi"]


@define
class MMPretrainDistributedTrainConfig(BaseConfigClass):
    """
    Definition of attributes needed to start a distributed training on multiple GPUs.

    It is leaned against the documentation of mmpretrain:
    https://mmpretrain.readthedocs.io/en/latest/user_guides/train.html

    TODO: Add this to the arc42 Documentation
    Launch multiple jobs simultaneously
    If you would like to launch multiple jobs on a single machine, e.g., 2 jobs of 4-GPU training
    on a machine with 8 GPUs, you need to specify different ports (29500 by default) for each job
    to avoid communication conflict.

    If you use dist_train.sh to launch training jobs, you can set the port in the commands.

    CUDA_VISIBLE_DEVICES=0,1,2,3 PORT=29500 ./tools/dist_train.sh ${CONFIG_FILE} 4
    CUDA_VISIBLE_DEVICES=4,5,6,7 PORT=29501 ./tools/dist_train.sh ${CONFIG_FILE} 4
    """

    __related_strict__ = True

    # Number of nodes, or the range of nodes in form <minimum_nodes>:<maximum_nodes>
    nnodes: str = related.StringField(default="1:1")

    # Number of workers per node; supported values: [auto, cpu, gpu, int]
    nproc_per_node = related.StringField(default="auto")

    # Rank of the node for multi-node distributed training.
    node_rank: int = related.IntegerField(default=0)

    # Synchronisation port for interprocess communication
    master_port: int = related.IntegerField(default=29500)

    # Synchronisation URI for interprocess communication
    master_address: str = related.StringField(default="localhost")


@define
class MMPretrainTrainDatasetConfig(BaseConfigClass):
    """Definition of annotation handler configuration that are used to define
    annotation data that is used in the training of a mmpretrain model. These
    will be used to overwrite dataloader configurations that are part of the
    original mmpretrain configuration.
    """

    data_root_dir: str = related.StringField(default="")

    train_annotation_handler_config: Optional[AnnotationHandlerConfig] = (
        related.ChildField(cls=AnnotationHandlerConfig, default=None)
    )

    val_annotation_handler_config: Optional[AnnotationHandlerConfig] = (
        related.ChildField(cls=AnnotationHandlerConfig, default=None)
    )

    test_annotation_handler_config: Optional[AnnotationHandlerConfig] = (
        related.ChildField(cls=AnnotationHandlerConfig, default=None)
    )


@define
class MMPretrainTrainConfig(BaseConfigClass):
    """Training related configuration parameters."""

    __related_strict__ = True

    argparse_config: MMPretrainTrainArgparseConfig = related.ChildField(
        cls=MMPretrainTrainArgparseConfig
    )

    best_metric_name: Optional[str] = related.ChildField(
        cls=Optional[str], default=None
    )

    multi_gpu_config: Optional[MMPretrainDistributedTrainConfig] = related.ChildField(
        cls=MMPretrainDistributedTrainConfig,
        required=False,
    )

    dataset_config: Optional[MMPretrainTrainDatasetConfig] = related.ChildField(
        cls=MMPretrainTrainDatasetConfig, required=False, default=None
    )


@define
class MMPretrainInferenceConfig(InferenceConfig):
    """Inference related configuration parameters.

    REMARK: The config_path attribute is inherited from the InferenceConfig,
            but is not used anymore and therefore is set to None.
    """

    __related_strict__ = True

    config_path: Optional[str] = None

    device_string: Optional[str] = related.StringField(required=False, default=None)

    reduction_class_mapping: Optional[ReductionMappingConfig] = related.ChildField(
        cls=ReductionMappingConfig, required=False, default=None
    )

    def check_values(self) -> bool:
        if self.config_path is not None:
            raise ValueError(
                "configuration.inference_config.config_path is not implemented for this module "
                "please use mmconfig.config_path instead."
            )

        return 0.0 <= self.score_threshold <= 1.0


@define
class MMConfig(BaseConfigClass):
    """Configuration for encapsulating a mmpretrain configuration.
    This is either a configuration file or a dictionary.
    """

    __related_strict__ = True

    config_path: Optional[str] = related.ChildField(
        cls=str, default=None, required=False
    )

    config_dict: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )

    cfg_options: Optional[Dict[str, Any]] = related.ChildField(
        cls=dict, default=None, required=False
    )


@define
class MMPretrainConfig(ModelConfiguration):
    """Definition of attributes for a MMPretrainModel."""

    __related_strict__ = True

    class_mapping: ClassMappingConfig = related.ChildField(cls=ClassMappingConfig)

    inference_config: MMPretrainInferenceConfig = related.ChildField(
        cls=MMPretrainInferenceConfig
    )

    train_config: MMPretrainTrainConfig = related.ChildField(cls=MMPretrainTrainConfig)

    mm_config: MMConfig = related.ChildField(cls=MMConfig)
